#!/usr/bin/perl -w  # this path to Perl is your_unix_system-dependent: either /usr/bin/perl or /usr/local/bin/perl
# Author:  Dr. Ruming Li -- January 10, 2011
# Purpose: Check geneID uniqueness and oligo cross-hybridization and make remaining oligos
#          have unique oligo probes generated with melting temperatures (Tm) being calculated.
# Usage:   perl oligo-maker.pl gene_file oligo_file [output_file][all_other_options]
#------------------------------------------------------------------------------------
use strict;
use warnings;
$0=~s/^\.\///;
my $usage = "\nUsage:\n\n$0 [switches]\n\nSwitches:\n\n-f <input-orfs/genes-file>
-i <input-oligo-probes-file>\n-I <input-overlaps-file>\n-o <output-oligo-probes-file>
-s <output-stats-tags-file>\n-l <oligo-probe-length>\n-c <total-strand-concentration>
-m <molar-salt-concentration>\n-t <DNA-melting-temperature>\n-d <difference-between-Tms>
-h <program-usage-info>\n
Note: the spacing following a switch (-x) is optional; eg, either '-d2' or '-d 2' is acceptable.


Synopsis:


-f <input-orfs/genes-file>

   refers to a sequence file that is the purged target gene file output from running the program
   orf-purger-v.2.0.pl. This file is better off being renamed for the recognizable purposes such
   as 'purged-target-sequences.ffn'.

-i <input-oligo-probes-file>

   refers to an oligo file that is the probe file downloaded from a commercial software (eg, eArray).
   When a Web-based probe design is successfully completed, download that probe file (zip-formatted)
   to your local machine. Then open that zip file, find one packed file called 'MOST_tdt', and drag-
   and-drop it to a directory (or better placed together with the purged-target-sequences.ffn file).

-I <input-overlaps-file>

   refers to a sequence file of Tab-delimited text type (*.tdt), which stores those ORFs overlapping
   with other ORFs (being a subsequence of another gene sequence). These ORF sequences have no oligo
   constructible since every oligo-long segment along such a sequence falls within the longer gene
   sequence with which it overlaps. However an overlap sequence should be used to check if an oligo
   searched from that longer gene sequence won't fall within it. That is, that oligo is constructed
   such that it won't overlap even this discarded ORF, thereby minimizing a potential risk of cross-
   hybridization provided that an ORF dropped as an overlap could exist as a real gene. This overlap
   data input is recommended to achieve the highest quality of constructed oligos, with some limited
   room for searching for an optimal Tm and a possible cost of a reduced number of oligo probes. The
   default value is 'none' and any filename input enables its functionality.


   Examples of command line specifications issued
   ----------------------------------------------
   If input files are in the working directory from which the program is invoked:

   > oligo-maker-v.2.0.pl -fpurged-target-sequences.ffn -iMOST_tdt

   If input files are not in the working directory and they are located in the directory /home/user/:

   > oligo-maker-v.2.0.pl -f/home/user/purged-target-sequences.ffn -i/home/user/MOST_tdt

   If an overlap sequence file is needed to input:

   > oligo-maker-v.2.0.pl -fpurged-target-sequences.ffn -iMOST_tdt -Ioverlaps.tdt


-o <output-oligo-probes-file>

   This is the file that contains all oligo probes and other gene feature information output from
   running this program, which will be used as one of input data files to produce final biological
   probe group and replicate probe group.

-s <output-stats-tags-file>

   This is the file that contains data statistics and gene tagging records for your information.

-l <oligo-probe-length>

   refers to the length of an oligo probe that you want to design. The default is 60bp (60mer).

-c <total-strand-concentration>

   is the total strand concentration in molar units that is used in thermodynamic calculations for
   the melting temperature (Tm). This value is automatically computed by the program but can be
   user-specified (eg, 33.58e-12) if it is consistent with that of a commercial software adoption.

-m <molar-salt-concentration>

   is the salt concentration in molar units that is used in thermodynamic calculations for the
   melting temperature (Tm). This value is automatically computed by the program but can be user-
   specified (eg, 0.74 M) if it is consistent with that of a commercial software adoption.

-t <DNA-melting-temperature>

   is the melting temperature (Tm) at which a DNA double helix dissociates into single strands.
   In oligo probe design, it also refers to the temperature at which 50% of the oligonucleotide
   and its perfect complement are in duplex. The default value is 80 C for prokaryotic genomes.

-d <difference-between-Tms>

   is the difference between a calculated Tm and the default Tm (80) or a Tm specified in Option
   -t. This value defaults to 10 as the less stringent criterion in order to include potentially
   workable oligo probes in microarray experiments, with the least loss of target information.";

my ($d,$e,$i,$j,$k,$l,$m,$n,$ok,$s,@idx,@tag,@gid,@acn,@gsb,@loc,@prd,@seq,@oid,@olg,@field);
my ($b,$c,$r,$A,$B,$C,$D,$M,$S,$Tm,$TM,$SH,$SS,@tm,@sid,@olp);

# Defaults
my $orfFile="input_genes.ffn";
my $olgFile="MOST_tdt";
my $olpFile="none";
my $outFile="oligos.tdt";
my $staTags="oligo-stats-tags.txt";
my $dnaConc="auto";
my $saltCon="auto";
my $melTemp=80;
my $oligLen=60;
my $diffTms=10;

unless (@ARGV) {}
else {while ($_=shift) {
if (/^-[Idcimflhost]/) {
if ($_ eq "-f") {$orfFile=(defined(local $_=shift)?$_:"");} elsif (/^-f/) {($orfFile)=/-f(.+)/;}
if ($_ eq "-i") {$olgFile=(defined(local $_=shift)?$_:"");} elsif (/^-i/) {($olgFile)=/-i(.+)/;}
if ($_ eq "-I") {$olpFile=(defined(local $_=shift)?$_:"");} elsif (/^-I/) {($olpFile)=/-I(.+)/;}
if ($_ eq "-o") {$outFile=(defined(local $_=shift)?$_:"");} elsif (/^-o/) {($outFile)=/-o(.+)/;}
if ($_ eq "-s") {$staTags=(defined(local $_=shift)?$_:"");} elsif (/^-s/) {($staTags)=/-s(.+)/;}
if ($_ eq "-c") {$dnaConc=(defined(local $_=shift)?$_:"");} elsif (/^-c/) {($dnaConc)=/-c(.+)/;}
if ($_ eq "-m") {$saltCon=(defined(local $_=shift)?$_:0);} elsif (/^-m/) {($saltCon)=/-m(.+)/;}
if ($_ eq "-t") {$melTemp=(defined(local $_=shift)?$_:0);} elsif (/^-t/) {($melTemp)=/-t(.+)/;}
if ($_ eq "-l") {$oligLen=(defined(local $_=shift)?$_:0);} elsif (/^-l/) {($oligLen)=/-l(.+)/;}
if ($_ eq "-d") {$diffTms=(defined(local $_=shift)?$_:0);} elsif (/^-d/) {($diffTms)=/-d(.+)/;}
if ($_ eq "-h") {print "$usage\n";exit;}
} else {die "\nUnknown argument/switch $_\n$usage\n";}}}
if ($orfFile!~/^[\w\/.-]*\w$/) {die "$orfFile: Invalid file name!\n";}
if ($olgFile!~/^[\w\/.-]*\w$/) {die "$olgFile: Invalid file name!\n";}
if ($olpFile!~/^[\w\/.-]*\w$/) {die "$olpFile: Invalid file name!\n";}
if ($melTemp!~/^\d+$/) {die "$melTemp: Not numerical parameter!\n";}
if ($oligLen!~/^\d+$/) {die "$oligLen: Not numerical parameter!\n";}
if ($diffTms!~/^\d+$/) {die "$diffTms: Not numerical parameter!\n";}

print "\nThis program will run based on the following settings:\n
-f input_genes_file   = $orfFile
-i input_oligos_file  = $olgFile
-I input_overlap_file = $olpFile
-o output_oligos_file = $outFile
-s out_stats_tag_file = $staTags
-c dna_concentration  = $dnaConc
-m salt_concentration = $saltCon
-t melt_temperature   = $melTemp
-l oligo_probe_length = $oligLen
-d difference_in_Tms  = $diffTms\n
Type in <$0 -h> for help and make your own settings.\n\n";

if ($dnaConc eq "auto") {$dnaConc="0.0e-12";}
if ($saltCon eq "auto") {$saltCon=0.74;}
if ($dnaConc!~/^[\d\.e-]+$/i) {die "$dnaConc: Not exponential parameter!\n";}
if ($saltCon!~/^0?\.\d+$/) {die "$saltCon: Not numerical/decimal parameter!\n";}

do {
	print "Proceed to run ? [y/b/n]: ";
	$b = <STDIN>;
	if ($b=~/^n$/i) {exit;}
} until ($b=~/^[yb]$/i);
if ($b=~/b/i) {
	print "\nPlease type bg to run in the background.\n";
	kill 'TSTP', $$;
}

open (ORFFILE, "<$orfFile") or die "$orfFile: No such a file.\n";
open (OLGFILE, "<$olgFile") or die "$olgFile: No such a file.\n";
open (OUTFILE, ">$outFile") or die "$!";
open (STATAGS, ">$staTags") or die "$!";

# Read in oligo data
while (<OLGFILE>) {
	next if ($.==1);
	chomp;
	@field = split(/\t/, $_, 7);
	push(@oid,$field[0]);
	push(@olg,$field[2]);
	push(@tm,$field[5]);
	#push(@idx,-1); # array for indexing
}
close(OLGFILE);

print STATAGS "\nNumber of input oligos = ",scalar @olg,"\n";
print "\n\nNumber of input oligos = ",scalar @olg,"\n\n";

# NN dimer enthalpy values
my %DH = (
	AA => -7.9,
	AC => -8.4,
	AG => -7.8,
	AT => -7.2,
	CA => -8.5,
	CC => -8.0,
	CG => -10.6,
	CT => -7.8,
	GA => -8.2,
	GC => -9.8,
	GG => -8.0,
	GT => -8.4,
	TA => -7.2,
	TC => -8.2,
	TG => -8.5,
	TT => -7.9,
);
# NN dimer entropy values
my %DS = (
	AA => -22.2,
	AC => -22.4,
	AG => -21.0,
	AT => -20.4,
	CA => -22.7,
	CC => -19.9,
	CG => -27.2,
	CT => -21.0,
	GA => -22.2,
	GC => -24.4,
	GG => -19.9,
	GT => -22.4,
	TA => -21.3,
	TC => -22.2,
	TG => -22.7,
	TT => -22.2,
);

if ($dnaConc eq "0.0e-12" && $saltCon==0.74) {
print "Computing strand and salt concentrations for oligo Tm...\n\n";
$n=22;
$D=1e+8;
$e=$#olg;
$d=$oligLen-2;
foreach my $a (-75..-73) {
	$m=-$a*0.01; # descending salt conc.
	$l=1e+8; # set to max
	# integral part in $n (eg, 10 in 10.1) is held as last strand conc.
	$A=$n+2; # beg-strand conc.
	$n=$r=0; # initialize base
	$B=99;   # end-strand conc.
	# only integral part is taken for '..' range values
	# eg, (10.12..99.23) ==> (10..99)
	LINE: foreach my $b ($A..$B) {
	$c=$n+$b*0.1**$r."e-12";
	$k=0;
	for ($i=0; $i<=$e; $i++) {
		$SH=$SS=0;
		# Terminal corrections
		if ($olg[$i]=~/^[CG]/) {
			$SH += 0.1;
			$SS += -2.8;
		} elsif ($olg[$i]=~/^[AT]/) {
			$SH += 2.3;
			$SS += 4.1;
		}
		if ($olg[$i]=~/[CG]$/) {
			$SH += 0.1;
			$SS += -2.8;
		} elsif ($olg[$i]=~/[AT]$/) {
			$SH += 2.3;
			$SS += 4.1;
		}
		for ($j=0; $j<=$d; $j++) {
			$s=substr($olg[$i],$j,2);
			$SH += $DH{$s};
			$SS += $DS{$s};
		}
		$Tm = (1000*$SH)/($SS+1.987*log($c/4))-273.15+16.6*log($m)/log(10);
		$k += abs($Tm-$tm[$i]);
	}
	if ($k < $l) {
		$l=$k;
		$C=$c;
		$M=$m;
		print "D=$k\tM=$m\t  C=$c\n";
	} elsif ($r<8) { # recursively calculated to a precision of 8 decimal digits
		$l=1e+8; # reset to max for recursively initial comparisons
		$n=$n+($b-2)*0.1**$r++; # determine a start base of strand concentrations
		if ($r==1) {$A=1;$B=19;} # beg-end values are calculated just once
		goto LINE; # Enter into fine computation-search from coarse one
	} else {
		if ($l < $D) {
			$D=$l;
			$dnaConc=$C;
			$saltCon=$M;
		}
		last; # exit from inner loop when done with fine computation
	}
}}
print "\nSalt concentration = $saltCon\nTotal strand concentration = $dnaConc\nMinimum accumulated difference = $D\n";
print STATAGS "Salt concentration = $saltCon\nTotal strand concentration = $dnaConc\nMinimum accumulated difference = $D\n";
}

# Read in orf data by sequence block
local $/="\n>";
while (<ORFFILE>) {
	chomp; # trim the trailing "\n>"
	/>?(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.+)/;
	push(@gid,$1); # Grab gene ID from header
	push(@acn,$2); # Grab gene accession
	push(@gsb,$3); # Grab gene symbol
	push(@loc,$4); # Grab gene locus/coords
	push(@prd,$5); # Grab gene product name
	s/.+\n//; # trim the 1st line of sequence block
	s/\n//g;  # strip newlines off sequence block
	push(@seq,$_); # save sequence string
	push(@tag,-1); # array for tagging
}
close(ORFFILE);

print "\nNumber of input orfs = ",scalar @seq,"\n\n";
print STATAGS "Number of input orfs = ",scalar @seq,"\n\n";

$/="\n"; # reset to normal input record separator
$e=$#seq;
$m=$#olg;
# Check geneID uniqueness and mark matches
for ($i=0; $i<=$e; $i++) {
	for ($j=0; $j<=$m; $j++) {
		if ($gid[$i] eq $oid[$j]) {
			if ($tag[$i] > -1) {
				print "$oid[$j] not unique in oligo entries!\n";
				print STATAGS "$oid[$j] not unique in oligo entries!\n";
				exit;
			}
			$tag[$i]=$j; # mark orf having an oligo
			$idx[$j]=$i; # mark relational indices
		}
	}
	if ($tag[$i] < 0) {
		print "ORF $gid[$i] has no oligo produced from a commercial software.\n";
		print STATAGS "ORF $gid[$i] has no oligo produced from a commercial software.\n";
	}
}

# Check if an oligo overlaps all other orfs
$n=$ok=0;
for ($j=0; $j<=$m; $j++) {
	for ($i=0; $i<=$e; $i++) {
		if ($tag[$i]!= $j) {
			if ($seq[$i]=~/$olg[$j]/) {
				print "ORF $oid[$j] has an x-hyb oligo.\n";
				print STATAGS "ORF $oid[$j] has an x-hyb oligo.\n";
				$tag[$idx[$j]]=-2; # mark orf having x-hyb oligo
				$n += 1;
				last;
			}
		}
	}
	if ($i > $e) {
		$ok += 1;
	}
}

print "\nNumber of x-hyb oligos = $n\n";
print "Number of unique oligos = $ok\n";
print STATAGS "\nNumber of x-hyb oligos = $n\n";
print STATAGS "Number of unique oligos = $ok\n";

if ($olpFile ne "none") {
# Read in overlaps data
open (OLPFILE, "<$olpFile") or die "$olpFile: No such a file.\n";
while (<OLPFILE>) {
	chomp;
	@field = split(/\t/, $_, 2);
	push(@sid,$field[0]);
	push(@olp,$field[1]);
}
close(OLPFILE);
print STATAGS "Number of input overlaps = ",scalar @olp,"\n\n";
print "Number of input overlaps = ",scalar @olp,"\n\n";
} else {
	print "\n";
	print STATAGS "\n";
}

# Create oligos for orfs w/ Tm calculated
my $p;
$A=$B=$c=$n=$m=0;
for ($i=0; $i<=$e; $i++) {
	if ($tag[$i] > -1) {
		print OUTFILE "$gid[$i]\t$acn[$i]\t$gsb[$i]\t$olg[$tag[$i]]\t$tag[$i]\t$tm[$tag[$i]]\t$loc[$i]\t$prd[$i]\n";
	} else {
		print "Constructing oligo for ORF $gid[$i]...\n";
		if ($olpFile ne "none") {
			($p)=grep{$sid[$_]=~/^$gid[$i]$/} 0..$#sid;
			if (defined($p)) { # geneID-->overlap found
				# sid and gid exactly match and array index returns;
				# if not matched, a '($p)=undef' status is returned.
				$A += 1; # Count geneID found in the overlap file
			}
		}
		$l=length($seq[$i])-$oligLen;
		$D=$diffTms;
		$M=0;
		for ($j=0; $j<=$l; $j++) {
			$s=substr($seq[$i],$j,$oligLen); # extract oligo seq
			for ($k=0; $k<=$e; $k++) {
				if ($k != $i) {
					last if ($seq[$k]=~/$s/);
				}
			}
			if ($k > $e) { # has unique oligo
				if (defined($p)) {
					if ($olp[$p]=~/$s/) {
						goto NEXT; # if oligo falling within an overlap, get next candidate oligo.
					} elsif ($M==0) {
						$B += 1; # One-time count of oligo sequence not falling within its overlap
					}
				}
				# Has unique oligo and calculate Tm
				$SH=$SS=0;
				# Terminal corrections
				if ($s =~ /^[CG]/) {
					$SH += 0.1;
					$SS += -2.8;
				} elsif ($s =~ /^[AT]/) {
					$SH += 2.3;
					$SS += 4.1;
				}
				if ($s =~ /[CG]$/) {
					$SH += 0.1;
					$SS += -2.8;
				} elsif ($s =~ /[AT]$/) {
					$SH += 2.3;
					$SS += 4.1;
				}
				for ($r=0; $r<=$d; $r++) {
					$_=substr($s,$r,2);
					$SH += $DH{$_};
					$SS += $DS{$_};
				}
				$Tm = (1000*$SH)/($SS+1.987*log($dnaConc/4))-273.15+16.6*log($saltCon)/log(10);
				$M = abs($Tm-$melTemp);
				if ($M < $D) {$S=$s;$TM=$Tm;$D=$M;} # Get Tm closest to a melting temperature (eg, 80C)
				if ($M == 0) {last;} # An exactly desired Tm hits, bail out of loop.
			}
			NEXT:
		}
		if ($D < $diffTms) {
			$TM = sprintf("%.2f",$TM);
			print "ORF $gid[$i] has oligo created.\n";
			print OUTFILE "$gid[$i]\t$acn[$i]\t$gsb[$i]\t$S\t$tag[$i]\t$TM\t$loc[$i]\t$prd[$i]\n";
			$m += 1;
		} elsif ($M==0) {
			print STATAGS "ORF $gid[$i] has no unique oligo found.\n";
			print STATAGS "This ORF is discarded as all its oligos overlap other genes.\n";
			print "ORF $gid[$i] has no unique oligo found.\n";
			print "This ORF is discarded as all its oligos overlap other genes.\n";
			$c += 1;
		} else {
			print STATAGS "ORF $gid[$i] has no oligo with required Tm.\n";
			print STATAGS "This ORF is discarded as there is no oligo constructible.\n";
			print "ORF $gid[$i] has no oligo with required Tm.\n";
			print "This ORF is discarded as there is no oligo constructible.\n";
			$tag[$i]=-3; # mark this orf for visualization
			$n += 1;
		}
	}
}

if ($olpFile ne "none") {
	print "\nNumber of gene ids found and processed in overlaps (",scalar @olp,") = $A\n";
	print "Number of oligos not falling within overlap sequences = $B\n";
	print STATAGS "\nNumber of gene ids found and processed in overlaps (",scalar @olp,") = $A\n";
	print STATAGS "Number of oligos not falling within overlap sequences = $B\n";
}
print "\nNumber of oligos not found = $c\n";
print "Number of oligos without required Tm = $n\n";
print "Number of constructed unique oligos = $m\n";
print "Number of total output unique oligos = ",$m+$ok,"\n\n";
print STATAGS "\nNumber of oligos not found = $c\n";
print STATAGS "Number of oligos without required Tm = $n\n";
print STATAGS "Number of constructed unique oligos = $m\n";
print STATAGS "Number of total output unique oligos = ",$m+$ok,"\n";

close(OUTFILE);
close(STATAGS);
print "Done.\n\n";
if ($b=~/b/i) {
	print "Please type fg to return to the foreground and exit.\n";
}
